# This Python file uses the following encoding: utf-8
"""
This module provides functions for super-uniform sampling of the unit
hypercube. 'Super-uniform' in this context means that the obtained point
sample is more uniform than a random uniform sample, which is a desirable
property in many applications. After creation, the samples can be
transformed from the unit hypercube to arbitrary cuboids.
"""
import math
import random
import itertools

import numpy as np

from diversipy.distance import calc_euclidean_dist_matrix
from diversipy.distance import DistanceMatrixFunction



def unitcube(dimension):
    """Shortcut to generate a tuple of bounds of the unit hypercube."""
    assert dimension > 0
    return [0.0] * dimension, [1.0] * dimension



def scaled(points, from_cuboid, to_cuboid):
    """Linear transformation between arbitrary cuboids.

    This function does not check if the points are actually inside
    `from_cuboid`.

    Parameters
    ----------
    points : array_like
        2-D array of points.
    from_cuboid : tuple
        Contains lower and upper boundaries.
    to_cuboid : tuple
        Contains lower and upper boundaries.

    Returns
    -------
    scaled_points : numpy array
        A new array containing the scaled points.

    """
    # sanity checks
    min_bounds_to, max_bounds_to = to_cuboid
    assert len(min_bounds_to) == len(max_bounds_to)
    for min_bound_to, max_bound_to in zip(min_bounds_to, max_bounds_to):
        assert max_bound_to >= min_bound_to
    min_bounds_from, max_bounds_from = from_cuboid
    assert len(min_bounds_from) == len(max_bounds_from)
    for min_bound_from, max_bound_from in zip(min_bounds_from, max_bounds_from):
        assert max_bound_from >= min_bound_from
    # scale
    length_factors = (np.asarray(max_bounds_to) - min_bounds_to)
    length_factors /= (np.asarray(max_bounds_from) - min_bounds_from)
    scaled_points = (np.asarray(points) - min_bounds_from) * length_factors
    scaled_points += min_bounds_to
    return scaled_points



def grid(num_points, dimension):
    """Create conventional grid in unit hypercube.

    Also related to full factorial designs.

    Parameters
    ----------
    num_points : int
        The number of points to generate.
        ``num_points ** (1/dimension)`` must be integer.
    dimension : int
        The dimension of the space.

    Returns
    -------
    points : (`num_points`, `dimension`) numpy array

    """
    points_per_axis = int(num_points ** (1.0 / dimension))
    assert points_per_axis ** dimension == num_points
    possible_values = list(range(points_per_axis))
    divisor = points_per_axis - 1.0
    for i in range(points_per_axis):
        possible_values[i] /= divisor
    points = np.array(list(itertools.product(possible_values, repeat=dimension)))
    return points



def sukharev_grid(num_points, dimension):
    """Create Sukharev grid in unit hypercube.

    Special property of this grid is that points are not placed on the
    boundaries of the hypercube, but at centroids of the `num_points`
    subcells. This design offers optimal results for the covering radius
    regarding distances based on the max-norm.

    Parameters
    ----------
    num_points : int
        The number of points to generate.
        ``num_points ** (1/dimension)`` must be integer.
    dimension : int
        The dimension of the space.

    Returns
    -------
    points : (`num_points`, `dimension`) numpy array

    """
    points_per_axis = int(num_points ** (1.0 / dimension))
    assert points_per_axis ** dimension == num_points
    possible_values = [x + 0.5 for x in range(points_per_axis)]
    divisor = points_per_axis
    for i in range(points_per_axis):
        possible_values[i] /= divisor
    points = np.array(list(itertools.product(possible_values, repeat=dimension)))
    return points



def random_uniform(num_points, dimension):
    """Syntactic sugar for :func:`numpy.random.rand`."""
    return np.random.rand(num_points, dimension)



def halton(num_points, dimension, skip=0):
    """Generate a Halton point set.

    Quasirandom sequence using the default initialization with the first
    `dimension` prime numbers.

    Parameters
    ----------
    num_points : int
        The number of points to generate.
    dimension : int
        The dimension of the space.
    skip : int, optional
        The first `skip` points of the sequence will be left out.

    Returns
    -------
    points : (`num_points`, `dimension`) numpy array

    """
    def is_prime(number):
        number = float(number)
        if number % 2 == 0 and number != 2:
            return False
        for divisor in range(3, int(number ** 0.5) + 1, 2):
            if number % divisor == 0:
                return False
        return True

    def generator(index, base):
        result = 0
        f = 1.0 / base
        i = index
        while i > 0:
            result += f * (i % base)
            i = math.floor(i / base)
            f /= base
        return result

    bases = [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59,
             61, 67, 71, 73, 79, 83, 89, 97, 101, 103, 107, 109, 113, 127, 131,
             137, 139, 149, 151, 157, 163, 167, 173, 179, 181, 191, 193, 197,
             199, 211, 223, 227, 229, 233, 239, 241, 251]
    while len(bases) < dimension:
        next_candidate = bases[-1] + 2
        while not is_prime(next_candidate):
            next_candidate += 2
        bases.append(next_candidate)
    points = []
    for i in range(skip, skip + num_points):
        point = [generator(i, bases[j]) for j in range(dimension)]
        points.append(point)
    return np.array(points)



def random_k_means(num_points,
                   dimension,
                   num_steps=None,
                   initial_points=None,
                   dist_matrix_function=None,
                   callback=None):
    """MacQueen's method.

    In its default setup, this algorithm converges to a centroidal Voronoi
    tesselation of the unit hypercube. Further information is given in
    [MacQueen1967]_.

    Parameters
    ----------
    num_points : int
        The number of points to generate.
    dimension : int
        The dimension of the space.
    num_steps : int, optional
        The number of iterations to carry out. Default is
        ``100 * num_points``.
    initial_points : array_like, optional
        The point set to improve (if None, a random uniform sample is
        drawn).
    dist_matrix_function : callable, optional
        The function to compute the distances. Default is Euclidean
        distance.
    callback : callable, optional
        If provided, it is called in each iteration with the current point
        set as argument for monitoring progress.

    Returns
    -------
    cluster_centers : (`num_points`, `dimension`) numpy array

    References
    ----------
    .. [MacQueen1967] MacQueen, J. Some methods for classification and
        analysis of multivariate observations. Proceedings of the Fifth
        Berkeley Symposium on Mathematical Statistics and Probability,
        Volume 1: Statistics, pp. 281--297, University of California Press,
        Berkeley, Calif., 1967.
        http://projecteuclid.org/euclid.bsmsp/1200512992.

    """
    # initialization
    if num_steps is None:
        num_steps = 100 * num_points
    if initial_points is None:
        cluster_centers = np.random.rand(num_points, dimension)
    elif len(initial_points) == num_points:
        cluster_centers = np.array(initial_points)
    else:
        raise ValueError("len(initial_points) must be equal to num_points")
    weights = [1.0] * num_points
    if dist_matrix_function is None:
        dist_matrix_function = calc_euclidean_dist_matrix
    # begin iteration
    for _ in range(num_steps):
        if callback is not None:
            callback(cluster_centers)
        random_point = np.random.rand(1, dimension)
        distances = dist_matrix_function(random_point, cluster_centers)
        random_point = random_point.ravel()
        nearest_index = np.argmin(distances, axis=1)
        nearest_cluster_center = cluster_centers[nearest_index, :].ravel()
        if hasattr(dist_matrix_function, "max_dists_per_dim") and dist_matrix_function.max_dists_per_dim is not None:
            one_dim_dists = np.abs(nearest_cluster_center - random_point)
            virtual_point = np.array(random_point)
            for j, dist in enumerate(one_dim_dists):
                if dist > 0.5:
                    if nearest_cluster_center[j] < 0.5:
                        virtual_point[j] = 1.0 - random_point[j]
                    else:
                        virtual_point[j] = 1.0 + random_point[j]
        else:
            virtual_point = random_point
        weight = weights[nearest_index]
        cluster_centers[nearest_index, :] = (weight * nearest_cluster_center + virtual_point) / (weight + 1.0)
        cluster_centers[nearest_index, :] %= 1.0
        assert np.all(cluster_centers[nearest_index, :] <= 1.0)
        assert np.all(cluster_centers[nearest_index, :] >= 0.0)
        weights[nearest_index] += 1.0
    return cluster_centers



def maximin_reconstruction(num_points,
                           dimension,
                           num_steps=None,
                           initial_points=None,
                           existing_points=None,
                           use_reflection_edge_correction=False,
                           dist_matrix_function=None,
                           callback=None):
    """Maximize the minimal distance in the unit hypercube with extensions.

    This algorithm carries out a user-specified number of iterations to
    maximize the minimal distance of a point in the set to 1) other points
    in the set, 2) existing (fixed) points, and 3) the boundary of the
    hypercube. Details can be found in [Wessing2015]_.

    Parameters
    ----------
    num_points : int
        The number of points to generate.
    dimension : int
        The dimension of the space.
    num_steps : int, optional
        The number of iterations to carry out. Default is
        ``100 * num_points``.
    initial_points : array_like, optional
        The point set to improve (if None, a random uniform sample is
        drawn).
    existing_points : array_like, optional
        Points that cannot be modified anymore, but should be considered in
        the distance computations.
    use_reflection_edge_correction : bool, optional
        If True, selection pressure in boundary regions will be increased by
        considering additional distances to virtual points, which are
        created by mirroring the real points at the boundary.
    dist_matrix_function : callable, optional
        The function to compute the distances. Default is Manhattan distance
        on a torus.
    callback : callable, optional
        If provided, it is called in each iteration with the current point
        set as argument for monitoring progress.

    Returns
    -------
    points : (`num_points`, `dimension`) numpy array

    References
    ----------
    .. [Wessing2015] Wessing, Simon (2015). Two-stage Methods for Multimodal
        Optimization. PhD Thesis, Technische Universität Dortmund.
        http://hdl.handle.net/2003/34148

    """
    def dist_to_bound(point):
        return min(point.min(), (1.0 - point).min())

    # initialization and sanity checks
    assert num_points > 0
    if num_steps is None:
        num_steps = 100 * num_points
    if initial_points is None:
        points = np.random.rand(num_points, dimension)
    elif len(initial_points) == num_points:
        points = np.array(initial_points)
        assert np.all(points >= 0.0)
        assert np.all(points <= 1.0)
    else:
        raise ValueError("len(initial_points) must be equal to num_points")
    if existing_points is None:
        existing_points = []
    existing_points = np.atleast_2d(existing_points)
    if existing_points.size == 0:
        num_existing_points = 0
    else:
        num_existing_points = len(existing_points)
    if dist_matrix_function is None:
        dist_matrix_function = DistanceMatrixFunction(exponent=1,
                                                      max_dists_per_dim=[1.0] * dimension)
    norm_of_one_vector = 2.0 * dist_matrix_function(np.atleast_2d([0.0] * dimension),
                                                    np.atleast_2d([0.5] * dimension))[0, 0]
    remaining_indices = list(range(num_points))
    random.shuffle(remaining_indices)
    removal_candidate_index = remaining_indices.pop()
    removal_candidate = points[removal_candidate_index]
    # initial distance calculations
    distances = dist_matrix_function(np.atleast_2d(removal_candidate), points)[0]
    distances[removal_candidate_index] = np.inf
    current_dist = distances.min()
    if num_existing_points > 0:
        dists_to_existing_points = dist_matrix_function(np.atleast_2d(removal_candidate),
                                                        existing_points)[0]
        current_dist = min(current_dist, dists_to_existing_points.min())
    if use_reflection_edge_correction:
        # compare with 2 * ||1|| * (distance to nearest boundary)
        relaxed_boundary_dist = 2.0 * norm_of_one_vector * dist_to_bound(removal_candidate)
        current_dist = min(current_dist, relaxed_boundary_dist)
    # maximize minimal distance
    for _ in range(num_steps):
        if callback is not None:
            callback(points)
        new_point = np.random.rand(1, dimension)
        new_dist = np.inf
        if use_reflection_edge_correction:
            # compare with 2 * ||1|| * (distance to nearest boundary)
            relaxed_boundary_dist = 2.0 * norm_of_one_vector * dist_to_bound(new_point)
            new_dist = relaxed_boundary_dist
        if new_dist >= current_dist:
            distances = dist_matrix_function(new_point, points)[0]
            distances[removal_candidate_index] = np.inf
            new_dist = min(new_dist, distances.min())
            if new_dist >= current_dist and num_existing_points > 0:
                dists_to_existing_points = dist_matrix_function(new_point, existing_points)[0]
                new_dist = min(new_dist, dists_to_existing_points.min())
        if new_dist >= current_dist:
            # accept new point
            points[removal_candidate_index] = new_point
            current_dist = new_dist
            removal_candidate = new_point
            # removal_candidate_index stays the same, but reset other indices
            remaining_indices = list(range(num_points))
            remaining_indices.remove(removal_candidate_index)
            random.shuffle(remaining_indices)
        else:
            # failed to find better point in this iteration
            if remaining_indices:
                # carry out one attempt to find new removal candidate
                removal_candidate_candidate_index = remaining_indices.pop()
                removal_candidate_candidate = points[removal_candidate_candidate_index]
                # calculate minimal distance
                distances = dist_matrix_function(np.atleast_2d(removal_candidate_candidate), points)[0]
                distances[removal_candidate_candidate_index] = np.inf
                candidate_candidate_dist = distances.min()
                if num_existing_points > 0:
                    dists_to_existing_points = dist_matrix_function(np.atleast_2d(removal_candidate_candidate),
                                                                    existing_points)[0]
                    candidate_candidate_dist = min(candidate_candidate_dist,
                                                   dists_to_existing_points.min())
                if use_reflection_edge_correction:
                    # compare with 2 * ||1|| * (distance to nearest boundary)
                    relaxed_boundary_dist = 2.0 * norm_of_one_vector
                    relaxed_boundary_dist *= dist_to_bound(removal_candidate_candidate)
                    candidate_candidate_dist = min(candidate_candidate_dist,
                                                   relaxed_boundary_dist)
                if candidate_candidate_dist <= current_dist:
                    # found new removal candidate
                    removal_candidate = removal_candidate_candidate
                    current_dist = candidate_candidate_dist
                    removal_candidate_index = removal_candidate_candidate_index
    return points



def conventional_stratified_sampling(num_points,
                                     dimension,
                                     full_output=False,
                                     bates_param=1):
    """Stratified sampling in the unit hypercube.

    This algorithm divides the hypercube into `num_points` subcells and
    draws a random uniform point from each cell. Thus, the result is
    stochastic, but more uniform than a random uniform sample. For further
    information see [McKay1979]_.

    Parameters
    ----------
    num_points : int
        The number of points to generate.
        ``num_points ** (1/dimension)`` must be integer.
    dimension : int
        The dimension of the space.
    full_output : bool
        If True, also the strata are returned as second argument. Default
        is False.
    bates_param : int, optional
        Each coordinate of a point sampled in a stratum is determined as
        the mean of this number of independent random uniform variables.
        Thus, the coordinates follow the Bates distribution.

    Returns
    -------
    points : (`num_points`, `dimension`) numpy array

    """
    # shortcuts & initialization
    assert num_points > 0
    assert dimension > 0
    points_per_axis = int(num_points ** (1.0 / dimension))
    if points_per_axis ** dimension != num_points:
        points_per_axis += 1
    assert points_per_axis ** dimension == num_points
    points = np.empty((num_points, dimension))
    final_strata = []
    extent = 1.0 / points_per_axis
    possible_values = [float(x) / points_per_axis for x in range(points_per_axis)]
    # calculate grid cells and sample
    for j, lower_bounds in enumerate(itertools.product(possible_values, repeat=dimension)):
        upper_bounds = [lower + extent for lower in lower_bounds]
        point = []
        for low_value, high_value in zip(lower_bounds, upper_bounds):
            if math.isinf(bates_param):
                point.append((low_value + high_value) * 0.5)
            else:
                uniform_sum = sum(random.uniform(low_value, high_value) for _ in range(bates_param))
                point.append(uniform_sum / bates_param)
        points[j] = point
        final_strata.append((list(lower_bounds), upper_bounds))
    if full_output:
        return points, final_strata
    else:
        return points



def stratified_sampling(num_points,
                        dimension,
                        cuboid=None,
                        full_output=False,
                        detect_special_case=True,
                        bates_param=1,
                        avoid_odd_numbers=True):
    """Generalized stratified sampling.

    Parameters
    ----------
    num_points : int
        The number of points to be sampled. An arbitrary number of points
        is possible for this algorithm.
    dimension : int
        The dimension of the search space.
    cuboid : tuple of list, optional
        Optionally specify the hypercube to be sampled in. If None, the
        unit hypercube is chosen.
    full_output : bool, optional
        If True, also the strata are returned as second argument.
    detect_special_case : bool, optional
        If True, ``num_points ** (1/dimension)`` is integer, and we are
        sampling the unit cube, use the original stratified sampling.
    bates_param : int, optional
        Each coordinate of a point sampled in a stratum is determined as
        the mean of this number of independent random uniform variables.
        Thus, the coordinates follow the Bates distribution.
    avoid_odd_numbers : bool, optional
        If this value is True, splits are chosen so that the resulting
        numbers are even, whenever possible. E.g., if a stratum with six
        points is splitted, it is not split into three and three, but two
        and four points.

    Returns
    -------
    points : (`num_points`, `dimension`) numpy array

    """
    # sanity checks
    assert num_points > 0
    assert dimension > 0
    if cuboid is None:
        cuboid = unitcube(dimension)
    assert len(cuboid[0]) == dimension
    assert len(cuboid[1]) == dimension
    assert bates_param > 0
    # special case
    if detect_special_case and cuboid == unitcube(dimension):
        points_per_axis = int(num_points ** (1.0 / dimension))
        is_integer_power = points_per_axis ** dimension == num_points
        # check because of potential rounding error:
        is_integer_power |= (points_per_axis + 1) ** dimension == num_points
        if is_integer_power:
            return conventional_stratified_sampling(num_points,
                                                    dimension,
                                                    full_output,
                                                    bates_param)
    # more initialization
    dimensions = list(range(dimension))
    points = np.empty((num_points, dimension))
    final_strata = []
    remaining_strata = [(num_points, cuboid)]
    # begin partitioning
    while remaining_strata:
        current_num_points, current_bounds = remaining_strata.pop()
        if current_num_points == 1:
            final_strata.append(current_bounds)
            continue
        current_lower, current_upper = current_bounds
        diffs = [current_upper[i] - current_lower[i] for i in dimensions]
        max_extent = max(diffs)
        max_extent_dims = [i for i in dimensions if diffs[i] == max_extent]
        num1 = int(current_num_points * 0.5)
        do_subtract_one = avoid_odd_numbers and current_num_points >= 6
        do_subtract_one &= num1 % 2 != 0 and current_num_points % 2 == 0
        if do_subtract_one:
            num1 -= 1
        num2 = current_num_points - num1
        if random.random() < 0.5:
            num1, num2 = num2, num1
        split_dim = random.choice(max_extent_dims)
        split_pos = float(num1) / current_num_points
        split_pos = current_lower[split_dim] + max_extent * split_pos
        new_upper = current_upper[:]
        new_upper[split_dim] = split_pos
        new_lower = current_lower[:]
        new_lower[split_dim] = split_pos
        stratum1 = (num1, (current_lower, new_upper))
        stratum2 = (num2, (new_lower, current_upper))
        remaining_strata.extend((stratum1, stratum2))
    # sample
    for j, stratum in enumerate(final_strata):
        lower, upper = stratum
        point = []
        for low_value, high_value in zip(lower, upper):
            if math.isinf(bates_param):
                point.append((low_value + high_value) * 0.5)
            else:
                uniform_sum = sum(random.uniform(low_value, high_value) for _ in range(bates_param))
                point.append(uniform_sum / bates_param)
        points[j] = point
    if full_output:
        return points, final_strata
    else:
        return points



def lhd_matrix(num_points, dimension):
    """Generate a random latin hypercube design matrix.

    Latin hypercube designs sometimes give an advantage over random uniform
    samples due to their perfect uniformity of one-dimensional projections.
    For further information see [McKay1979]_. This algorithm has linear run
    time.

    Parameters
    ----------
    num_points : int
        The number of points to generate.
    dimension : int
        The dimension of the space.

    Returns
    -------
    design : (`num_points`, `dimension`) numpy array
        Contains integers corresponding to the bins of a virtual grid.

    References
    ----------
    .. [McKay1979] McKay, M.D.; Beckman, R.J.; Conover, W.J. (May 1979).
        A Comparison of Three Methods for Selecting Values of Input
        Variables in the Analysis of Output from a Computer Code.
        Technometrics 21(2): 239-245.

    """
    design = np.zeros((num_points, dimension), dtype="i")
    permutation = np.random.permutation
    for i in range(dimension):
        design[:, i] = permutation(np.arange(0, num_points))
    return design



def improved_lhd_matrix(num_points,
                        dimension,
                        num_candidates=100,
                        target_value=None,
                        dist_matrix_function=None):
    """Generate an 'improved' latin hypercube design matrix.

    This implementation uses an algorithm with quadratic run time. It is a
    greedy construction heuristic starting with a randomly chosen point.
    In each iteration, a number of random candidates is evaluated by a
    criterion that considers a candidate's distance to the previously chosen
    points. The best point according to this criterion is included in the
    LHD. The concept originally stems from [Beachkofski2002]_. The algorithm
    implemented here was proposed in [Wessing2015]_.

    Parameters
    ----------
    num_points : int
        The number of points to generate.
    dimension : int
        The dimension of the space.
    num_candidates : int, optional
        The number of random candidates considered for every point to be
        added to the LHD.
    target_value : float, optional
        The distance a candidate should ideally have to the already chosen
        points of the LHD.
    dist_matrix_function : callable, optional
        Defines the distance used. Default is Manhattan distance on a torus
        (maximum distance is ``num_points - 1``, well-suited for
        :func:`edge_lhs`).

    Returns
    -------
    design : (`num_points`, `dimension`) numpy array
        Contains integers corresponding to the bins of a virtual grid.

    References
    ----------
    .. [Beachkofski2002] Beachkofski, B.; Grandhi, R. (2002). Improved
        Distributed Hypercube Sampling. American Institute of Aeronautics
        and Astronautics Paper 1274.

    """
    # shortcuts & initialization
    dimensions = list(range(dimension))
    permutation = np.random.permutation
    if dist_matrix_function is None:
        max_dists = [num_points - 1.0] * dimension
        dist_matrix_function = DistanceMatrixFunction(exponent=1, max_dists_per_dim=max_dists)
    if target_value is None:
        target_value = np.inf
    design = np.empty((num_points, dimension), dtype="i")
    available_bins = []
    first_point = []
    bins = range(num_points)
    for _ in dimensions:
        dim_bins = list(bins)
        random.shuffle(dim_bins)
        first_point.append(dim_bins.pop())
        available_bins.append(dim_bins)
    design[0, :] = first_point
    # begin choosing other points
    for i in range(1, num_points - 1):
        # create a set of num_candidates random points obeying the LHD property
        available_bins_array = np.array(available_bins).T
        num_available = float(num_points - i)
        duplication_factor = int(math.ceil(num_candidates / num_available))
        candidates = np.vstack([available_bins_array] * duplication_factor)
        for dim in dimensions:
            candidates[:, dim] = permutation(candidates[:, dim])
        # calculate distances to already chosen points
        distances = dist_matrix_function(design[0:i, :], candidates[0:num_candidates, :])
        min_dists = distances.min(axis=0)
        # select best point according to distance criterion
        if not math.isinf(target_value):
            selected_index = np.argmin(np.abs(min_dists - target_value))
        else:
            selected_index = np.argmax(min_dists)
        selected_point = candidates[selected_index, :]
        design[i, :] = selected_point
        # maintain bin data structure
        for value, avail in zip(selected_point, available_bins):
            avail.remove(value)
    last_point = [avail.pop() for avail in available_bins]
    for avail in available_bins:
        assert not avail  # should be empty
    design[-1, :] = last_point
    return design



def has_lhd_property(design_matrix):
    """Check if design matrix has the LHD property.

    It is assumed that counting starts from zero and points are arranged
    row-wise. So, each column must consist of a permutation of
    ``range(num_points)``.

    Parameters
    ----------
    design_matrix : array_like
        2-D array of integer coordinates.

    Returns
    -------
    is_lhd : bool

    """
    design_matrix = np.asarray(design_matrix)
    num_points, dimension = design_matrix.shape
    required_numbers = list(range(num_points))
    for i in range(dimension):
        ith_column = design_matrix[:, i].tolist()
        if required_numbers != sorted(ith_column):
            return False
    return True



def perturbed_lhs(design_matrix):
    """Transform a design matrix into a sample in the unit hypercube.

    Applies random perturbations so that each point is distributed randomly
    uniform in its grid cell. This is the variant proposed by [McKay1979]_.
    It is not checked if `design_matrix` actually is a LHD.

    Parameters
    ----------
    design_matrix : array_like
        Array containing integers to indicate the bins occupied by each
        point.

    Returns
    -------
    points : (`num_points`, `dimension`) numpy array

    """
    design_matrix = np.asarray(design_matrix)
    num_points, num_dimensions = design_matrix.shape
    # make a copy
    points = design_matrix + np.random.rand(num_points, num_dimensions)
    points /= float(num_points)
    return points



def centered_lhs(design_matrix):
    """Transform a design matrix into a sample in the unit hypercube.

    Each point is placed at the centroid of a subcell in the assumed grid
    over the cube. It is not checked if `design_matrix` actually is a LHD.

    Parameters
    ----------
    design_matrix : array_like
        Array containing integers to indicate the bins occupied by each
        point.

    Returns
    -------
    points : (`num_points`, `dimension`) numpy array

    """
    design_matrix = np.asarray(design_matrix)
    num_points, _ = design_matrix.shape
    # make a copy
    points = design_matrix + 0.5
    points /= float(num_points)
    return points



def edge_lhs(design_matrix):
    """Transform a design matrix into a sample in the unit hypercube.

    The transformation is so that each face of the hypercube is sampled by
    exactly one point. Use this transformation if you want to maximize the
    minimal distance between points in the design. It is not checked if
    `design_matrix` actually is a LHD.

    Parameters
    ----------
    design_matrix : array_like
        Array containing integers to indicate the bins occupied by each
        point.

    Returns
    -------
    points : (`num_points`, `dimension`) numpy array

    """
    design_matrix = np.asarray(design_matrix)
    num_points, _ = design_matrix.shape
    # make a copy
    points = np.array(design_matrix, dtype="d")
    points /= (num_points - 1.0)
    return points



# This Python file uses the following encoding: utf-8
"""
This module contains several functions to measure diversity and a few
related concepts. The diversity indicators all have different advantages and
disadvantages. An overview is given in [Wessing2015]_.

"""
import math

import numpy as np
from scipy.spatial import Voronoi

from diversipy.hycusampling import stratified_sampling
from diversipy.distance import calc_euclidean_dist_matrix
from diversipy.distance import calc_dists_to_boundary


def covering_radius(points, repair_margin=1e-8, full_output=False):
    """Calculate the covering radius of points in the unit hypercube.

    The indicator is calculated for Euclidean distance via a Voronoi
    tessellation. It should be minimized. Coordinates of points must be
    >0 and <1.

    .. note:: This function requires SciPy for the Voronoi tessellation.
    .. warning:: The run time is exponential in the dimension.

    Parameters
    ----------
    points : array_like
        2-D data structure holding the points.
    repair_margin : float, optional
        Due to the inherent inaccuracy of floating point arithmetic, some
        vertices of the Voronoi tessellation that should be exactly on a
        boundary may be located outside the hypercube. This parameter
        specifies a margin around the hypercube, whose points are repaired
        and included in the distance calculations. The higher this value,
        the more expensive and the more reliable are the calculations.
    full_output : bool, optional
        If true, also the Voronoi tessellation itself is returned.

    Returns
    -------
    cov_radius : float
        The covering radius of the point set.
    voronoi_tessellation : scipy.spatial.qhull.Voronoi
        Data structure containing the Voronoi tessellation produced by the
        Qhull library. See SciPy documentation for details.

    """
    points = np.asarray(points)
    if len(points) < 1:
        return float("inf")
    assert np.all(points >= 0.0)
    assert np.all(points <= 1.0)
    _, dimension = points.shape
    mirrored_points = [points]
    for i in range(dimension):
        points_copy = np.array(points)
        points_copy[:, i] *= -1
        mirrored_points.append(points_copy)
        points_copy = np.array(points)
        points_copy[:, i] = 2.0 - points[:, i]
        mirrored_points.append(points_copy)
    mirrored_points = np.vstack(mirrored_points)
    voronoi_tessellation = Voronoi(mirrored_points, qhull_options="Q12")
    vertices = voronoi_tessellation.vertices
    lower_bounds_satisfied = (vertices >= -repair_margin).sum(axis=1)
    upper_bounds_satisfied = (vertices <= 1.0 + repair_margin).sum(axis=1)
    total_bounds_satisfied = lower_bounds_satisfied + upper_bounds_satisfied
    relevant_indices = np.where(total_bounds_satisfied == 2 * dimension)[0]
    relevant_vertices = vertices[relevant_indices]
    relevant_vertices = np.minimum(relevant_vertices, 1.0)
    relevant_vertices = np.maximum(relevant_vertices, 0.0)
    dist_matrix = calc_euclidean_dist_matrix(relevant_vertices, points)
    cov_radius = dist_matrix.min(axis=1).max()
    if full_output:
        return cov_radius, voronoi_tessellation
    else:
        return cov_radius



def covering_radius_ub(points, strata, dist_matrix_function=None):
    """Upper bound for the covering radius.

    Parameters
    ----------
    points : array_like
        The points to assess.
    strata : sequence of tuple
        A partition of the unit hypercube as produced by
        :func:`stratified_sampling<diversipy.hycusampling.stratified_sampling>`
        with `full_output` enabled. Each stratum must contain exactly one
        point.
    dist_matrix_function : callable, optional
        An arbitrary distance function. Default is Euclidean distance.

    Returns
    -------
    cr_ub : float
        Upper bound of the covering radius for the point set.

    """
    points = np.asarray(points)
    if len(points) < 1:
        return float("inf")
    assert np.all(points >= 0.0)
    assert np.all(points <= 1.0)
    assert len(points) == len(strata)
    if dist_matrix_function is None:
        dist_matrix_function = calc_euclidean_dist_matrix
    cr_ub = 0.0
    for point, stratum in zip(points, strata):
        furthest_corner = []
        for i in range(len(point)):
            if point[i] - stratum[0][i] > stratum[1][i] - point[i]:
                furthest_corner.append(stratum[0][i])
            else:
                furthest_corner.append(stratum[1][i])
        furthest_corner = np.atleast_2d(furthest_corner)
        dist = dist_matrix_function(np.atleast_2d(point), furthest_corner)[0, 0]
        if dist > cr_ub:
            cr_ub = dist
    return cr_ub



def covering_radius_lb(points,
                       num_monte_carlo_points,
                       block_size=10000,
                       dist_matrix_function=None):
    """Monte Carlo lower bound for the covering radius.

    Parameters
    ----------
    points : array_like
        The points to assess.
    num_monte_carlo_points : int
        The number of points used in the estimation of the covering radius.
        Higher values lead to a better approximation quality. Points are
        drawn by stratified sampling in blocks of `block_size`.
    block_size : int, optional
        The Monte Carlo points are drawn in blocks of this size to avoid
        exceeding the memory capacity.
    dist_matrix_function : callable, optional
        An arbitrary distance function. Default is Euclidean distance.

    Returns
    -------
    cr_lb : float
        Lower bound of the covering radius for the point set.

    """
    points = np.asarray(points)
    if len(points) < 1:
        return float("inf")
    assert np.all(points >= 0.0)
    assert np.all(points <= 1.0)
    _, dim = points.shape
    if dist_matrix_function is None:
        dist_matrix_function = calc_euclidean_dist_matrix
    monte_carlo_points = stratified_sampling(num_monte_carlo_points, dim)
    i = 0
    cr_lb = 0
    while i < len(monte_carlo_points):
        dist_matrix = dist_matrix_function(monte_carlo_points[i:(i + block_size)],
                                           points)
        i += block_size
        cr_lb = max(cr_lb, dist_matrix.min(axis=1).max())
    return cr_lb



def solow_polasky_diversity(points, activity_param=1.0, dist_matrix_function=None):
    """Calculate the Solow-Polasky diversity for a set of points.

    This diversity indicator was introduced in [Solow1994]_ and is to be
    maximized. The algorithm has cubic run time, because the pseudoinverse
    of a correlation matrix has to be computed. The assumed correlation
    function is ``exp(-activity_param * dist)``.

    Parameters
    ----------
    points : array_like
        2-D data structure holding the points.
    activity_param : float, optional
        Parameter controlling the strength of correlation. It must hold
        ``0 < activity_param <= 2``. Default is 1.
    dist_matrix_function : callable, optional
        A metric distance function. Default is Euclidean distance.

    Returns
    -------
    diversity : float

    References
    ----------
    .. [Solow1994]  Solow, Andrew R.; Polasky, Stephen (1994). Measuring
        biological diversity. Environmental and Ecological Statistics,
        Vol. 1, No. 2, pp. 95-103. https://dx.doi.org/10.1007/BF02426650

    """
    if len(points) == 0:
        return 0.0
    if dist_matrix_function is None:
        dist_matrix_function = calc_euclidean_dist_matrix
    points = np.asarray(points)
    dist_matrix = dist_matrix_function(points, points)
    correlation_matrix = np.exp(-activity_param * dist_matrix)
    try:
        # compute pseudoinverse
        inverse_matrix = np.linalg.pinv(correlation_matrix)
    except np.linalg.linalg.LinAlgError:
        return 0.0
    return inverse_matrix.sum()



def weitzman_diversity(points, dist_matrix_function=None):
    """Calculate the Weitzman diversity for a set of points.

    This diversity indicator was introduced in [Weitzman1992]_. It is to be
    maximized.

    .. warning:: This implementation has exponential run time in the number
                 of points!

    Parameters
    ----------
    points : array_like
        2-D data structure holding the points.
    dist_matrix_function : callable, optional
        An arbitrary distance function. Default is Euclidean distance.

    Returns
    -------
    diversity : float

    References
    ----------
    .. [Weitzman1992] Weitzman, Martin L. (May 1992). On Diversity. The
        Quarterly Journal of Economics, Vol. 107, No. 2, pp. 363-405
        https://www.jstor.org/stable/2118476

    """
    def diversity_recursive(indices, dist_matrix):
        num_points = len(indices)
        if num_points == 1:
            # end of recursion
            return 1.0
        # find minimal distance
        min_dist = float("inf")
        min_dist_pair = None, None
        for i in range(num_points):
            index_i = indices[i]
            for j in range(i + 1, num_points):
                distance = dist_matrix[index_i, indices[j]]
                if distance < min_dist:
                    min_dist = distance
                    min_dist_pair = (i, j)
        a, b = min_dist_pair
        # prepare for recursive calls
        indices_without_a = indices[:]
        del indices_without_a[a]
        indices_without_b = indices[:]
        del indices_without_b[b]
        diversity_without_a = diversity_recursive(indices_without_a, dist_matrix)
        diversity_without_b = diversity_recursive(indices_without_b, dist_matrix)
        if diversity_without_a > diversity_without_b:
            return min_dist + diversity_without_a
        else:
            return min_dist + diversity_without_b

    num_points = len(points)
    if num_points == 0:
        return 0.0
    if dist_matrix_function is None:
        dist_matrix_function = calc_euclidean_dist_matrix
    points = np.asarray(points)
    dist_matrix = dist_matrix_function(points, points)
    return diversity_recursive(list(range(num_points)), dist_matrix)



def sum_of_dists(points, dist_matrix_function=None):
    """Calculate the square root of the sum of all pairwise distances.

    This indicator is to be maximized.

    .. warning:: This function is only included here for comparisons. It is
        actually not well suited as diversity indicator, as explained in
        [Solow1994]_.

    Parameters
    ----------
    points : array_like
        2-D data structure holding the points.
    dist_matrix_function : callable, optional
        An arbitrary distance function. Default is Euclidean distance.

    Returns
    -------
    spread : float

    """
    num_points = len(points)
    if num_points == 0:
        return 0.0
    if dist_matrix_function is None:
        dist_matrix_function = calc_euclidean_dist_matrix
    points = np.asarray(points)
    dist_matrix = dist_matrix_function(points, points)
    spread = math.sqrt(dist_matrix.sum() * 0.5)
    return spread



def average_inverse_dist(points, exponent=None, max_dist=1.0, dist_matrix_function=None):
    """Calculate the average inverse distance.

    For each pair of points, the value ``(max_dist / dist) ** exponent`` is
    computed. The average of all these values is the indicator value, which
    is to be minimized.

    Parameters
    ----------
    points : array_like
        2-D data structure holding the points.
    exponent : int or float, optional
        Exponent in the calculations explained above. Default is
        ``dimension + 1``.
    max_dist : float, optional
        Maximally possible distance or an arbitrary constant.
    dist_matrix_function : callable, optional
        An arbitrary distance function. Default is Euclidean distance.

    Returns
    -------
    diversity : float

    """
    num_points = len(points)
    if num_points < 2:
        return 0.0
    points = np.asarray(points)
    num_points, dimension = points.shape
    if exponent is None:
        exponent = dimension + 1
    if dist_matrix_function is None:
        dist_matrix_function = calc_euclidean_dist_matrix
    dist_matrix = dist_matrix_function(points, points)
    sum_of_inv_dists = 0.0
    num_zeros = 0
    for i in range(num_points):
        for j in range(i + 1, num_points):
            distance = dist_matrix[i, j]
            if distance == 0.0:
                num_zeros += 1
            else:
                sum_of_inv_dists += (max_dist / distance) ** exponent
    if num_zeros > 0:
        return float("inf")
    else:
        num_dists = num_points * (num_points - 1) / 2.0
        return sum_of_inv_dists ** (1.0 / exponent) / num_dists



def separation_dist(points, dist_matrix_function=None):
    """Calculate the minimal pairwise distance.

    This indicator is to be maximized.

    Parameters
    ----------
    points : array_like
        2-D data structure holding the points.
    dist_matrix_function : callable, optional
        An arbitrary distance function. Default is Euclidean distance.

    Returns
    -------
    min_dist : float

    """
    num_points = len(points)
    if num_points < 2:
        return 0.0
    points = np.asarray(points)
    if dist_matrix_function is None:
        dist_matrix_function = calc_euclidean_dist_matrix
    dist_matrix = dist_matrix_function(points, points)
    for i in range(num_points):
        dist_matrix[i, i] = np.inf
    min_dist = dist_matrix.min()
    return min_dist



def sum_of_nn_dists(points, dist_matrix_function=None):
    """Calculate the sum of nearest-neighbor distances

    This indicator is to be maximized.

    Parameters
    ----------
    points : array_like
        2-D data structure holding the points.
    dist_matrix_function : callable, optional
        An arbitrary distance function. Default is Euclidean distance.

    Returns
    -------
    sum_of_nn_dists : float

    """
    num_points = len(points)
    if num_points < 2:
        return 0.0
    points = np.asarray(points)
    if dist_matrix_function is None:
        dist_matrix_function = calc_euclidean_dist_matrix
    dist_matrix = dist_matrix_function(points, points)
    for i in range(num_points):
        dist_matrix[i, i] = np.inf
    nn_dists = dist_matrix.min(axis=0)
    return nn_dists.sum()







def unanchored_L2_discrepancy(points):
    """Calculate unanchored L2 discrepancy.

    Discrepancy is to be minimized. Note that the square root is already
    taken. Coordinates of points must be >=0 and <=1. Run time is quadratic.
    For details see [Morokoff1994]_.

    Parameters
    ----------
    points : array_like
        2-D data structure holding the points.

    Returns
    -------
    discrepancy : float

    References
    ----------
    .. [Morokoff1994] Morokoff, William J.; Caflisch, Russel E. (1994).
        Quasi-random sequences and their discrepancies. SIAM Journal on
        Scientific Computing, Vol. 15, No. 6, pp. 1251-1279.
        https://dx.doi.org/10.1137/0915077

    """
    if len(points) == 0:
        return 0.0
    points = np.asarray(points)
    assert np.all(points >= 0.0)
    assert np.all(points <= 1.0)
    num_points, dimension = points.shape
    num_batches = int(math.ceil(num_points / 1000.0))
    part1_sum = 0.0
    for j in range(num_batches):
        lo1 = j * 1000
        hi1 = (j + 1) * 1000
        for k in range(j, num_batches):
            lo2 = k * 1000
            hi2 = (k + 1) * 1000
            part1_matrix = (1.0 - np.maximum(points[lo1:hi1, 0, None], points[lo2:hi2, 0]))
            part1_matrix *= np.minimum(points[lo1:hi1, 0, None], points[lo2:hi2, 0])
            for i in range(1, dimension):
                part1_matrix *= (1.0 - np.maximum(points[lo1:hi1, i, None], points[lo2:hi2, i]))
                part1_matrix *= np.minimum(points[lo1:hi1, i, None], points[lo2:hi2, i])
            batch_sum = part1_matrix.sum()
            part1_sum += batch_sum
            if j != k:
                part1_sum += batch_sum
    del part1_matrix
    part2_sum = np.sum((points * (1.0 - points)).prod(axis=1))
    result = part1_sum / num_points ** 2.0
    result -= part2_sum * (2.0 ** (1.0 - dimension) / num_points)
    result += 12 ** -dimension
    return math.sqrt(result)



def expected_unanchored_L2_discrepancy(num_points, dimension):
    """Expected value for unanchored L2 discrepancy of random uniform points.

    Note that this is the square root of :math:`\\mathrm{E}(T^2)`, i.e.
    ``sqrt(1.0 / n * (6 ** -d) * (1 - 2 ** -d))``. For details see
    [Morokoff1994]_.

    """
    assert num_points > 0
    assert dimension > 0
    return math.sqrt(1.0 / num_points * (6 ** -dimension) * (1 - 2 ** -dimension))



def mean_dist_to_boundary(points):
    """Calculate the mean distance to the boundary of this point set."""
    if len(points) == 0:
        return 0.0
    points = np.asarray(points)
    assert np.all(points >= 0.0)
    assert np.all(points <= 1.0)
    dists = calc_dists_to_boundary(points)
    return np.mean(dists)



def expected_dist_to_boundary(dimension):
    """The expected distance to the boundary for random uniform points."""
    assert dimension > 0
    return 0.5 / (1 + dimension)



def averaged_hausdorff_dist(points1, points2, exponent=1, dist_matrix_function=None):
    """Calculate the averaged Hausdorff distance between two point sets.

    As defined in the paper [Schuetze2012]_.

    Parameters
    ----------
    points1 : array_like
        2-D data structure holding the first set.
    points2 : array_like
        2-D data structure holding the second set.
    exponent : int, optional
        An exponent to control the penalization of outliers (the higher the
        larger the exponent is).
    dist_matrix_function : callable, optional
        An arbitrary distance function. Default is Euclidean distance.

    Returns
    -------
    ahd : float

    References
    ----------
    .. [Schuetze2012] Schütze, O.; Esquivel, X.; Lara, A.; Coello Coello,
        Carlos A. (2012). Using the Averaged Hausdorff Distance as a
        Performance Measure in Evolutionary Multiobjective Optimization.
        IEEE Transactions on Evolutionary Computation, Vol.16, No.4,
        pp. 504-522. https://dx.doi.org/10.1109/TEVC.2011.2161872

    """
    if math.isinf(exponent):
        return hausdorff_dist(points1, points2, dist_matrix_function)
    points1 = np.asarray(points1)
    num_points, dimension = points1.shape
    assert num_points > 0 and dimension > 0
    points2 = np.asarray(points2)
    num_points, dimension = points2.shape
    assert num_points > 0 and dimension > 0
    if dist_matrix_function is None:
        dist_matrix_function = calc_euclidean_dist_matrix
    dist_matrix = dist_matrix_function(points1, points2)
    min_dists1 = dist_matrix.min(axis=0)
    min_dists2 = dist_matrix.min(axis=1)
    part1 = (min_dists1 ** exponent).sum() / len(min_dists1) ** (1.0 / exponent)
    part2 = (min_dists2 ** exponent).sum() / len(min_dists2) ** (1.0 / exponent)
    ahd = max(part1, part2)
    return ahd



def hausdorff_dist(points1, points2, dist_matrix_function=None):
    """Calculate the Hausdorff distance between two point sets.

    Parameters
    ----------
    points1 : array_like
        2-D data structure holding the first set.
    points2 : array_like
        2-D data structure holding the second set.
    dist_matrix_function : callable, optional
        An arbitrary distance function. Default is Euclidean distance.

    Returns
    -------
    hd_dist : float

    """
    points1 = np.asarray(points1)
    num_points, dimension = points1.shape
    assert num_points > 0 and dimension > 0
    points2 = np.asarray(points2)
    num_points, dimension = points2.shape
    assert num_points > 0 and dimension > 0
    if dist_matrix_function is None:
        dist_matrix_function = calc_euclidean_dist_matrix
    dist_matrix = dist_matrix_function(points1, points2)
    min_dists1 = dist_matrix.min(axis=0)
    min_dists2 = dist_matrix.min(axis=1)
    hd_dist = max(min_dists1.max(), min_dists2.max())
    return hd_dist



"""
diversipy is a collection of algorithms for sampling in hypercubes,
selecting diverse subsets, and measuring diversity.

For further documentation see the submodules.
"""

from diversipy.hycusampling import *
from diversipy.subset import *
from diversipy.indicator import *
from diversipy.distance import *


__version__ = "0.6"



"""
This module provides some basic distance measures and a class to instantiate
arbitrary distance functions based on Minkowski metrics. For more exotic
distance functions (and faster implementations) have a look at
:mod:`scipy.spatial.distance`.
"""
import math

import numpy as np



def calc_dists_to_boundary(points, cuboid=None):
    """Calculate the distance of each point to the boundary of some cuboid.

    This distance is simply the minimum of all differences between
    a point and the lower and upper bounds. This function also checks if all
    calculated distances are larger than zero. If not, some points must be
    located outside the cuboid.

    Parameters
    ----------
    points : array_like
        2-D array of `n` points.
    cuboid : tuple of array_like, optional
        Contains the min and max bounds of the considered cuboid. If
        omitted, the unit hypercube is assumed.

    Returns
    -------
    distances : numpy array
        1-D array of `n` distances

    """
    if cuboid is None:
        dists_to_min_bounds = points.min(axis=1)
        dists_to_max_bounds = (1.0 - points).min(axis=1)
    else:
        min_bounds, max_bounds = cuboid
        dists_to_min_bounds = (points - np.asarray(min_bounds)).min(axis=1)
        dists_to_max_bounds = (np.asarray(max_bounds) - points).min(axis=1)
    distances = np.minimum(dists_to_min_bounds, dists_to_max_bounds)
    assert np.all(distances >= 0.0)  # are all points contained in cuboid?
    return distances



def calc_euclidean_dist_matrix(points1, points2):
    """Calculate Euclidean distance matrix between `points1` and `points2`.

    Generates one column of the matrix at a time.

    Parameters
    ----------
    points1 : array_like
        2-D array of n points.
    points2 : array_like
        2-D array of m points.

    Returns
    -------
    distances : (n,m) numpy array

    """
    num_points1, dim1 = points1.shape
    num_points2, dim2 = points2.shape
    assert dim1 == dim2
    distances = np.zeros((num_points1, num_points2))
    for i in range(dim1):
        diff2 = points1[:, i, None] - points2[:, i]
        diff2 **= 2
        distances += diff2
    # in place sqrt
    np.sqrt(distances, distances)
    return distances



def calc_manhattan_dist_matrix(points1, points2):
    """Calculate Manhattan distance matrix between `points1` and `points2`.

    Generates one column of the matrix at a time.

    Parameters
    ----------
    points1 : array_like
        2-D array of n points.
    points2 : array_like
        2-D array of m points.

    Returns
    -------
    distances : (n,m) numpy array

    """
    num_points1, dim1 = points1.shape
    num_points2, dim2 = points2.shape
    assert dim1 == dim2
    distances = np.zeros((num_points1, num_points2))
    for i in range(dim1):
        diff = np.abs(points1[:, i, None] - points2[:, i])
        distances += diff
    return distances



class DistanceMatrixFunction:
    """General distance function.

    This distance function can handle arbitrary exponents and can optionally
    calculate torus distances. Slightly slower than the specialized
    versions. Special cases ``exponent = 1`` and ``exponent = 2`` correspond
    to Manhattan and Euclidean distance, respectively.

    """
    def __init__(self, exponent=2, max_dists_per_dim=None):
        """Constructor.

        Parameters
        ----------
        exponent : scalar, optional
            The exponent in the distance calculation.
        max_dists_per_dim : array_like, optional
            1-D array of largest possible distance in each dimension.
            Providing these values has the consequence of treating the
            cuboid as a torus. This is useful for eliminating edge effects
            induced by the lack of neighbor points outside the bounds of
            the cuboid.

        """
        self.exponent = exponent
        self.max_dists_per_dim = max_dists_per_dim
        if max_dists_per_dim is not None:
            for max_dist in max_dists_per_dim:
                assert max_dist > 0 or max_dist is None


    def __call__(self, points1, points2):
        """Calculate distance matrix, one column at a time."""
        exponent = self.exponent
        max_dists_per_dim = self.max_dists_per_dim
        num_points1, dim1 = points1.shape
        num_points2, dim2 = points2.shape
        assert dim1 == dim2
        distances = np.zeros((num_points1, num_points2))
        if math.isinf(exponent):
            for i in range(dim1):
                diff = np.abs(points1[:, i, None] - points2[:, i], dtype="d")
                if max_dists_per_dim is not None:
                    np.minimum(max_dists_per_dim[i] - diff, diff, out=diff)
                assert np.all(diff >= 0)
                distances = np.maximum(diff, distances)
        else:
            for i in range(dim1):
                diff = np.abs(points1[:, i, None] - points2[:, i], dtype="d")
                if max_dists_per_dim is not None:
                    np.minimum(max_dists_per_dim[i] - diff, diff, out=diff)
                assert np.all(diff >= 0)
                diff **= exponent
                distances += diff
            # in place root function
            np.power(distances, 1.0 / exponent, out=distances)
        return distances



# This Python file uses the following encoding: utf-8
"""
This module contains algorithms for the task of subset selection: suppose
you have a set of points in :math:`\\mathbb{R}^n` and want to select a sample
of them distributed as uniform as possible. This problem is related to
clustering, with the difference that when using clustering, you usually want
to retain the structure of the original point set.
"""
import heapq
import random

import numpy as np

from diversipy.distance import calc_dists_to_boundary
from diversipy.distance import calc_euclidean_dist_matrix



class MinBoundingBox:
    """Data structure containing some information about a cluster of points.

    The box does not store the points themselves, but merely a reference
    to a container of a (possibly) larger set of points and the indices of
    points in this container belonging to this box. Most importantly, the
    box holds cached information about the minimal and maximal coordinates
    of the points and the largest difference between them. This information
    is important for the part-and-select algorithm (PSA) [Salomon2013]_.
    Comparison operators are overloaded to compare by this value for easy
    use in a binary heap.

    """

    def __init__(self, all_points, member_indices):
        self.sort_key = None
        self.dim_index = None
        self.min_bounds = None
        self.max_bounds = None
        self.all_points = all_points
        self.member_indices = member_indices
        # fill in information
        self.update_information()


    def update_information(self):
        """Gather information from the points and save it."""
        all_points = self.all_points
        member_indices = self.member_indices
        dimension = len(all_points[0])
        members_array = np.take(all_points, member_indices, axis=0)
        min_bounds = members_array.min(axis=0)
        max_bounds = members_array.max(axis=0)
        # determine dimension with largest difference
        sort_key = float("inf")
        dim_index = None
        for j in range(dimension):
            neg_diff = min_bounds[j] - max_bounds[j]
            if neg_diff < sort_key:
                sort_key = neg_diff
                dim_index = j
        # save data
        self.sort_key = sort_key
        self.dim_index = dim_index
        self.min_bounds = min_bounds
        self.max_bounds = max_bounds


    def closest_point_index(self, sought_point,
                            candidate_indices,
                            dist_matrix_function=None):
        """Return the index of the point closest to a certain target.

        Parameters
        ----------
        sought_point : array_like
            The target point.
        candidate_indices : array_like
            Indices into the container of all points for the candidates to
            be considered.
        dist_matrix_function : callable, optional
            An arbitrary distance function. Default is Euclidean distance.

        Returns
        -------
        closest_point_index : integer

        """
        all_points = self.all_points
        if dist_matrix_function is None:
            dist_matrix_function = calc_euclidean_dist_matrix
        candidates_array = np.take(all_points, candidate_indices, axis=0)
        distances = dist_matrix_function(np.atleast_2d(sought_point),
                                         candidates_array)
        closest_point_index = candidate_indices[np.argmin(distances)]
        return closest_point_index


    def obtain_representative(self, selection_target="centroid_of_hypercube",
                              tournament_size=0,
                              dist_matrix_function=None):
        """Return a point representing this cluster.

        Parameters
        ----------
        selection_target : string, optional
            Indicates which strategy is used to determine a representative.
            This must be one of ('random_uniform', 'centroid_of_hypercube',
            'center_of_mass', 'max_dist_from_boundary').
        tournament_size : int, optional
            Optionally restrict the candidates for selection to
            ``0 < tournament_size < len(member_indices)`` randomly chosen
            points. Any other value of this parameter leads to consideration
            of all points. Note that a value of 1 leads to a random choice
            among the clusters members, regardless of the target point.
        dist_matrix_function : callable, optional
            An arbitrary distance function. Default is Euclidean distance.
            This choice does not affect the selection strategy
            ``max_dist_from_boundary``.

        Returns
        -------
        representative : array_like

        """
        index = self.obtain_representative_index(selection_target,
                                                 tournament_size,
                                                 dist_matrix_function)
        representative = self.all_points[index]
        return representative


    def obtain_representative_index(self,
                                    selection_target="centroid_of_hypercube",
                                    tournament_size=0,
                                    dist_matrix_function=None):
        """Return the index to a point representing this cluster.

        Parameters
        ----------
        selection_target : string, optional
            Indicates which strategy is used to determine a representative.
            This must be one of ('random_uniform', 'centroid_of_hypercube',
            'center_of_mass', 'max_dist_from_boundary').
        tournament_size : int, optional
            Optionally restrict the candidates for selection to
            ``0 < tournament_size < len(member_indices)`` randomly chosen
            points. Any other value of this parameter leads to consideration
            of all points. Note that a value of 1 leads to a random choice
            among the clusters members, regardless of the target point.
        dist_matrix_function : callable, optional
            An arbitrary distance function. Default is Euclidean distance.
            This choice does not affect the selection strategy
            ``max_dist_from_boundary``.

        Returns
        -------
        representative_index : int

        """
        all_points = self.all_points
        member_indices = self.member_indices
        min_bounds = self.min_bounds
        max_bounds = self.max_bounds
        dimension = len(min_bounds)
        # determine target point
        if selection_target == "random_uniform":
            # random point in the space (!) of the bounding box
            target_point = [random.uniform(min_bounds[i], max_bounds[i]) for i in range(dimension)]
        elif selection_target == "center_of_mass":
            members_array = np.take(all_points, member_indices, axis=0)
            target_point = np.mean(members_array, axis=0)
        elif selection_target == "centroid_of_hypercube" or selection_target == "max_dist_from_boundary":
            target_point = [(max_bounds[i] + min_bounds[i]) * 0.5 for i in range(dimension)]
        else:
            raise ValueError("Unknown target mode '" + selection_target + "'")
        # determine subset of points as candidates
        if tournament_size <= 0 or tournament_size >= len(member_indices):
            candidate_point_indices = member_indices
        else:
            candidate_point_indices = random.sample(member_indices,
                                                    tournament_size)
        # find best of those according to selection criterion
        if selection_target == "max_dist_from_boundary":
            hypercube = (min_bounds, max_bounds)
            candidates_array = np.take(all_points, candidate_point_indices, axis=0)
            dists_to_bound = calc_dists_to_boundary(candidates_array, hypercube)
            max_dist_index = np.argmax(dists_to_bound)
            index = candidate_point_indices[max_dist_index]
        else:
            index = self.closest_point_index(target_point,
                                             candidate_point_indices,
                                             dist_matrix_function)
        return index


    def __lt__(self, other):
        return self.sort_key < other.sort_key


    def __gt__(self, other):
        return self.sort_key > other.sort_key


    def __eq__(self, other):
        return self.sort_key == other.sort_key


    def __le__(self, other):
        return self.sort_key <= other.sort_key


    def __ge__(self, other):
        return self.sort_key >= other.sort_key


    def __ne__(self, other):
        return self.sort_key != other.sort_key



def psa_partition(points, num_clusters, available_points_indices=None):
    """Partition the data set into the given number of clusters.

    The approach was originally proposed in [Salomon2013]_. The
    implementation here is the slightly improved version from
    [Wessing2015]_.

    .. note:: This algorithm fails if there are less than `num_clusters`
        distinct points in the original set.

    Parameters
    ----------
    points : array_like
        2-D data structure holding the points.
    num_clusters : int
        The number of 'clusters' to be identified.
    available_points_indices : array_like, optional
        If desired, the set of considered points can already be restricted
        here by only providing a subset of the indices into `points`.

    Returns
    -------
    clusters : list of MinBoundingBox

    References
    ----------
    .. [Salomon2013] Salomon, Shaul; Avigad, Gideon; Goldvard, Alex;
        Schütze, Oliver (2013). PSA -- A New Scalable Space Partition
        Based Selection Algorithm for MOEAs. EVOLVE -- A Bridge between
        Probability, Set Oriented Numerics, and Evolutionary Computation II,
        Advances in Intelligent Systems and Computing, Springer, Vol. 175,
        pp. 137-151. https://dx.doi.org/10.1007/978-3-642-31519-0_9

    """
    if available_points_indices is None:
        available_points_indices = list(range(len(points)))
    assert num_clusters <= len(available_points_indices)
    assert num_clusters > 0
    clusters = []
    most_dissimilar_cluster = MinBoundingBox(points, available_points_indices)
    while len(clusters) + 1 < num_clusters:
        split_index = most_dissimilar_cluster.dim_index
        lower_bound = most_dissimilar_cluster.min_bounds[split_index]
        upper_bound = most_dissimilar_cluster.max_bounds[split_index]
        split_position = (lower_bound + upper_bound) / 2.0
        indices1 = []
        indices2 = []
        for member_index in most_dissimilar_cluster.member_indices:
            if points[member_index][split_index] < split_position:
                indices1.append(member_index)
            else:
                indices2.append(member_index)
        cluster1 = MinBoundingBox(points, indices1)
        cluster2 = MinBoundingBox(points, indices2)
        heapq.heappush(clusters, cluster1)
        most_dissimilar_cluster = heapq.heappushpop(clusters, cluster2)
    heapq.heappush(clusters, most_dissimilar_cluster)
    return clusters



def psa_select(points,
               num_selected_points,
               available_points_indices=None,
               selection_target="centroid_of_hypercube",
               tournament_size=0,
               dist_matrix_function=None):
    """Combine partitioning and determination of representatives.

    The approach was originally proposed in [Salomon2013]_. The
    implementation here is the slightly improved version from
    [Wessing2015]_.

    Parameters
    ----------
    points : array_like
        2-D data structure holding the points.
    num_selected_points : int
        The number of points to be selected.
    available_points_indices : array_like, optional
        If desired, the set of considered points can already be restricted
        here by only providing a subset of the indices into `points`.
    selection_target : string, optional
        Indicates which strategy is used to determine a representative. This
        must be one of ('random_uniform', 'centroid_of_hypercube',
        'center_of_mass', 'max_dist_from_boundary').
    tournament_size : int, optional
        Optionally restrict the candidates for selection to
        ``0 < tournament_size < len(member_indices)`` randomly chosen
        points. Any other value of this parameter leads to consideration
        of all points. Note that a value of 1 leads to a random choice
        among the clusters members, regardless of the target point.
    dist_matrix_function : callable, optional
        An arbitrary distance function. Default is Euclidean distance.
        This choice does not affect the selection strategy
        ``max_dist_from_boundary``.

    Returns
    -------
    representatives : array_like

    """
    clusters = psa_partition(points, num_selected_points, available_points_indices)
    representatives = []
    for cluster in clusters:
        representative = cluster.obtain_representative(selection_target,
                                                       tournament_size,
                                                       dist_matrix_function)
        representatives.append(representative)
    if isinstance(points, np.ndarray):
        representatives = np.array(representatives)
    return representatives



def select_greedy_maximin(points,
                          num_selected_points,
                          existing_points=None,
                          dist_matrix_function=None):
    """Greedily select a subset according to maximin criterion.

    This selection approach corresponds to the indicator
    :func:`separation_dist() <diversipy.indicator.separation_dist>`.

    Parameters
    ----------
    points : array_like
        2-D data structure holding the points.
    num_selected_points : int
        The number of points to be selected.
    existing_points : array_like, optional
        Points that cannot be modified anymore, but should be considered in
        the distance computations.
    dist_matrix_function : callable, optional
        An arbitrary distance function. Default is Euclidean distance.

    Returns
    -------
    selected_points : array_like

    """
    assert num_selected_points <= len(points)
    assert num_selected_points > 0
    points_array = np.asarray(points)
    if existing_points is None or len(existing_points) == 0:
        existing_points = [random.choice(points)]
    existing_points = np.atleast_2d(existing_points)
    if existing_points.size == 0:
        existing_points = np.array([random.choice(points)])
    if dist_matrix_function is None:
        dist_matrix_function = calc_euclidean_dist_matrix
    num_points, _ = points_array.shape
    aggregated_dist_criteria = np.zeros(num_points) + np.inf
    distances = dist_matrix_function(existing_points, points_array)
    aggregated_dist_criteria = distances.min(axis=0)
    previous_index = np.argmax(aggregated_dist_criteria)
    selected_indices = [previous_index]
    while len(selected_indices) < num_selected_points:
        previous_point = np.atleast_2d(points_array[previous_index])
        distances = dist_matrix_function(previous_point, points_array)
        aggregated_dist_criteria = np.minimum(aggregated_dist_criteria,
                                              distances.ravel())
        previous_index = np.argmax(aggregated_dist_criteria)
        selected_indices.append(previous_index)
    if isinstance(points, list):
        return [points[i] for i in selected_indices]
    else:
        return np.take(points_array, selected_indices, axis=0)



def select_greedy_maxisum(points,
                          num_selected_points,
                          existing_points=None,
                          dist_matrix_function=None):
    """Greedily select a subset according to maxisum criterion.

    This selection approach corresponds to the indicator
    :func:`sum_of_dists() <diversipy.indicator.sum_of_dists>`.

    .. warning:: This function does not result in a uniformly distributed
        subset of the points, instead it selects extremal points.

    Parameters
    ----------
    points : array_like
        2-D data structure holding the points.
    num_selected_points : int
        The number of points to be selected.
    existing_points : array_like, optional
        Points that cannot be modified anymore, but should be considered in
        the distance computations.
    dist_matrix_function : callable, optional
        An arbitrary distance function. Default is Euclidean distance.

    Returns
    -------
    selected_points : array_like

    """
    assert num_selected_points <= len(points)
    assert num_selected_points > 0
    points_array = np.asarray(points)
    if existing_points is None or len(existing_points) == 0:
        existing_points = [random.choice(points)]
    existing_points = np.atleast_2d(existing_points)
    if existing_points.size == 0:
        existing_points = np.array([random.choice(points)])
    if dist_matrix_function is None:
        dist_matrix_function = calc_euclidean_dist_matrix
    num_points, _ = points_array.shape
    aggregated_dist_criteria = np.zeros(num_points)
    distances = dist_matrix_function(existing_points, points_array)
    aggregated_dist_criteria = distances.sum(axis=0)
    previous_index = np.argmax(aggregated_dist_criteria)
    selected_indices = [previous_index]
    while len(selected_indices) < num_selected_points:
        previous_point = np.atleast_2d(points[previous_index])
        distances = dist_matrix_function(previous_point, points_array)
        aggregated_dist_criteria += distances.ravel()
        aggregated_dist_criteria[selected_indices] = -np.inf
        previous_index = np.argmax(aggregated_dist_criteria)
        selected_indices.append(previous_index)
    if isinstance(points, list):
        return [points[i] for i in selected_indices]
    else:
        return np.take(points_array, selected_indices, axis=0)



def select_greedy_energy(points,
                         num_selected_points,
                         existing_points=None,
                         exponent=None,
                         dist_matrix_function=None):
    """Greedily select a subset according to potential energy.

    This selection approach uses the Riesz energy as selection criterion,
    which is related to the indicator
    :func:`average_inverse_dist() <diversipy.indicator.average_inverse_dist>`.
    The uniformity of the selected points can be controlled by `exponent`.
    More information about this criterion can for example be found in
    [Hardin2004]_.

    Parameters
    ----------
    points : array_like
        2-D data structure holding the points.
    num_selected_points : int
        The number of points to be selected.
    existing_points : array_like, optional
        Points that cannot be modified anymore, but should be considered in
        the distance computations.
    exponent : int or float, optional
        Parameter controlling the uniformity. Must be >= 0.
        Values greater or equal the dimension lead to uniformity, smaller
        values to a higher density in exterior regions.
    dist_matrix_function : callable, optional
        An arbitrary distance function. Default is Euclidean distance.

    Returns
    -------
    selected_points : array_like

    References
    ----------
    .. [Hardin2004] Hardin, Douglas P.; Saff, Ed B. (2004). Discretizing
        Manifolds via Minimum Energy Points. Notices of the American
        Mathematical Society, Vol. 51, No. 10, pp. 1186-1194.
        http://www.math.vanderbilt.edu/~hardin/papers/HSb2004.pdf

    """
    assert num_selected_points <= len(points)
    assert num_selected_points > 0
    points_array = np.asarray(points)
    if existing_points is None or len(existing_points) == 0:
        existing_points = [random.choice(points)]
    existing_points = np.atleast_2d(existing_points)
    if existing_points.size == 0:
        existing_points = np.array([random.choice(points)])
    if dist_matrix_function is None:
        dist_matrix_function = calc_euclidean_dist_matrix
    num_points, dimension = points_array.shape
    if exponent is None:
        exponent = dimension + 1
    aggregated_dist_criteria = np.zeros(num_points)
    distances = dist_matrix_function(existing_points, points_array)
    with np.errstate(divide='ignore'):
        if exponent == 0:
            distances = np.log(1.0 / distances)
        else:
            distances = 1.0 / (distances ** exponent)
    aggregated_dist_criteria = distances.sum(axis=0)
    previous_index = np.argmin(aggregated_dist_criteria)
    selected_indices = [previous_index]
    while len(selected_indices) < num_selected_points:
        previous_point = np.atleast_2d(points[previous_index])
        distances = dist_matrix_function(previous_point, points_array)
        with np.errstate(divide='ignore'):
            if exponent == 0:
                distances = np.log(1.0 / distances)
            else:
                distances = 1.0 / (distances ** exponent)
        aggregated_dist_criteria += distances.ravel()
        aggregated_dist_criteria[selected_indices] = np.inf
        previous_index = np.argmin(aggregated_dist_criteria)
        selected_indices.append(previous_index)
    if isinstance(points, list):
        return [points[i] for i in selected_indices]
    else:
        return np.take(points_array, selected_indices, axis=0)



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
import os
import io
import re


# for getting the __version__
def read(*names, **kwargs):
    with io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8")
    ) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


if __name__ == '__main__':
    here = os.path.abspath(os.path.dirname(__file__))

    # Get the long description from the relevant file
    with open(os.path.join(here, 'DESCRIPTION.rst'), encoding='utf-8') as f:
        long_description = f.read()

    name = "diversipy"

    setup(
        name=name,
        version=find_version(name, "__init__.py"),
        description='Sample in hypercubes, select diverse subsets, and measure diversity',
        long_description=long_description,
        url='https://ls11-www.cs.tu-dortmund.de/people/swessing/diversipy/doc/',
        author='Simon Wessing',
        author_email='simon.wessing@tu-dortmund.de',
        license='BSD',
        # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
        classifiers=[
            # How mature is this project? Common values are
            #   3 - Alpha
            #   4 - Beta
            #   5 - Production/Stable
            'Development Status :: 4 - Beta',

            "Intended Audience :: Science/Research",
            "Topic :: Scientific/Engineering",
            "Topic :: Scientific/Engineering :: Mathematics",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",

            'License :: OSI Approved :: BSD License',

            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.4',
        ],
        keywords='diversity uniformity discrepancy distance hypercube subset doe dace space-filling design experiments',
        #packages=["diversipy"],
        packages=find_packages(exclude=['test']),
        install_requires=['numpy', 'scipy'],
        requires=['numpy', 'scipy']
    )
